from operator import attrgetter
from typing import Dict, List, Iterable

import attr
from openpyxl import Workbook
from openpyxl.conftest import Worksheet

from .models import Style, CellWithSize
from .utils import str_cell_range, style_and_merge_cell_range, style_single_cell


def xlsx_from_json(json_data: Dict, default_style: Style = None) -> Workbook:
    default_style = default_style or Style()

    wb = Workbook()

    start_row = json_data.get("start_row", 1)
    start_column = json_data.get("start_column", 1)

    filler = SheetFiller(wb.active, start_column, start_row, default_style)
    filler.fill(json_data)

    return wb


@attr.s(auto_attribs=True)
class SheetFiller:
    sheet: Worksheet
    start_column: int
    start_row: int
    default_style: Style

    def fill(self, json_data: Dict) -> None:
        current_row = self.start_row

        for row_data in json_data["rows"]:
            cells = self._fill_row(current_row, row_data["cells"])
            row_height = max(*map(attrgetter("height"), cells), 1)
            current_row += row_height

    def _fill_row(self, row: int, cells_data: List[Dict]) -> Iterable[CellWithSize]:
        column = self.start_column

        for cell_data in cells_data:
            cell = self._create_cell(row, column, cell_data)
            column += cell.width
            yield cell

    def _create_cell(self, row: int, column: int, cell_data: Dict) -> CellWithSize:
        cell = self.sheet.cell(row, column)

        cell.value = cell_data["value"]

        width = cell_data.get("width", 1)
        height = cell_data.get("height", 1)

        style = Style.from_json(cell_data.get("style", {}), self.default_style)

        if width == 1 and height == 1:
            style_single_cell(cell, style)
        else:
            cell_range = str_cell_range(column, row, column + width, row + height)
            style_and_merge_cell_range(self.sheet, cell_range, style)

        return CellWithSize(cell, width, height)



import attr
from openpyxl.cell import Cell
from openpyxl.conftest import Worksheet
from openpyxl.styles import Border
from openpyxl.utils import get_column_letter

from .models import Style


def str_cell_range(start_column: int, start_row: int, end_column: int, end_row: int) -> str:
    from_column = get_column_letter(start_column)
    to_column = get_column_letter(end_column)
    return f"{from_column}{start_row}:{to_column}{end_row}"


def style_single_cell(cell: Cell, style: Style) -> None:
    for attrib, value in attr.asdict(style).items():
        if value:
            setattr(cell, attrib, value)


def style_and_merge_cell_range(ws: Worksheet, cell_range: str, style: Style) -> None:
    """
    Source:
    https://openpyxl.readthedocs.io/en/2.5/styles.html#styling-merged-cells
    """
    top = Border(top=style.border.top)
    left = Border(left=style.border.left)
    right = Border(right=style.border.right)
    bottom = Border(bottom=style.border.bottom)

    first_cell = ws[cell_range.split(":")[0]]
    if style.alignment:
        ws.merge_cells(cell_range)
        first_cell.alignment = style.alignment

    if style.font:
        first_cell.font = style.font

    rows = ws[cell_range]
    for cell in rows[0]:
        cell.border = cell.border + top
    for cell in rows[-1]:
        cell.border = cell.border + bottom

    for row in rows:
        l = row[0]
        r = row[-1]
        l.border = l.border + left
        r.border = r.border + right
        if style.fill:
            for c in row:
                c.fill = style.fill



from itertools import chain

import pytest
from openpyxl import Workbook
from openpyxl.cell import Cell
from openpyxl.conftest import Worksheet
from openpyxl.styles import Font

from xlsx_from_json import xlsx_from_json, Style


@pytest.fixture()
def default_style():
    return Style(font=Font(bold=True))


@pytest.fixture()
def json_data_with_single_cell():
    return {
        "rows": [
            {
                "cells": [
                    {
                        "value": "Sample text",
                        "style": {
                            "font": {
                                "name": "Times New Roman",
                                "size": 12
                            },
                            "border": {
                                "bottom": {
                                    "border_style": "medium",
                                    "color": "00000000"
                                }
                            },
                            "alignment": {
                                "horizontal": "center"
                            },
                            "fill": {
                                "patternType": "gray125"
                            }
                        }
                    }
                ]
            }
        ],
        "start_column": 3
    }


@pytest.fixture()
def json_data_with_sized_cell():
    return {
        "rows": [
            {
                "cells": [
                    {
                        "value": "Sample text",
                        "width": 5,
                        "height": 2,
                        "style": {
                            "font": {
                                "name": "Times New Roman",
                                "size": 12
                            },
                            "border": {
                                "bottom": {
                                    "border_style": "medium",
                                    "color": "FFFFFFFF"
                                }
                            }
                        }
                    }
                ]
            }
        ]
    }


@pytest.fixture()
def json_data_with_start_row_and_multiple_cells():
    return {
        "rows": [
            {
                "cells": [
                    {
                        "value": "1x2",
                        "height": 2
                    }
                ]
            },
            {
                "cells": [
                    {
                        "value": "2x1",
                        "width": 2
                    },
                    {
                        "value": "1x1"
                    }
                ]
            }
        ],
        "start_row": 2
    }


@pytest.fixture()
def workbook(json_data_with_single_cell, default_style) -> Workbook:
    return xlsx_from_json(json_data_with_single_cell, default_style)


@pytest.fixture()
def sheet(workbook) -> Worksheet:
    return workbook.active


@pytest.fixture()
def cell(sheet) -> Cell:
    return sheet.cell(row=1, column=3)


def test_sheet_has_values(sheet):
    assert list(filter(None, chain.from_iterable(sheet.values))) == ["Sample text"]


def test_cell_is_shifted_by_offset(cell):
    assert cell.column == 'C'
    assert cell.value == "Sample text"


def test_cell_has_font(cell):
    assert cell.font.name == "Times New Roman"
    assert cell.font.size == 12


def test_cell_has_font_bold_from_default_style(cell):
    assert cell.font.bold


def test_cell_has_border(workbook, cell):
    assert cell.border.bottom.border_style == "medium"
    assert cell.border.bottom.color.rgb == "00000000"


def test_cell_has_alignment(workbook, cell):
    assert cell.alignment.horizontal == "center"


def test_cell_has_fill(workbook, cell):
    assert cell.fill.patternType == "gray125"


def test_sized_cell_is_rendered_as_merged_cells_and_style_set(json_data_with_sized_cell):
    workbook: Workbook = xlsx_from_json(json_data_with_sized_cell)
    sheet = workbook.active
    assert sheet.cell(3, 6).border.bottom.border_style == "medium"


def test_sheet_fill_starts_with_start_row(json_data_with_start_row_and_multiple_cells):
    workbook: Workbook = xlsx_from_json(json_data_with_start_row_and_multiple_cells)
    sheet = workbook.active
    assert sheet.cell(row=2, column=1).value == "1x2"


def test_sheet_has_two_rows(json_data_with_start_row_and_multiple_cells):
    workbook: Workbook = xlsx_from_json(json_data_with_start_row_and_multiple_cells)
    sheet = workbook.active
    assert sheet.cell(row=2, column=1).value == "1x2"
    assert sheet.cell(row=4, column=1).value == "2x1"

def test_sheet_has_two_columns(json_data_with_start_row_and_multiple_cells):
    workbook: Workbook = xlsx_from_json(json_data_with_start_row_and_multiple_cells)
    sheet = workbook.active
    assert sheet.cell(row=4, column=1).value == "2x1"
    assert sheet.cell(row=4, column=3).value == "1x1"




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = [
    "attrs>=18.1.0",
    "openpyxl>=2.5.4"
]

setuptools.setup(
    name="xlsx_from_json",
    version="0.1.1",
    author="potykion",
    author_email="potykion@gmail.com",
    description="Creates xlsx from json via openpyxl.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/potykion/xlsx_from_json",
    install_requires=requirements,
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    include_package_data=True,
)



from typing import Dict

import attr
from openpyxl.cell import Cell
from openpyxl.styles import Font, Border, Fill, Alignment, PatternFill, Side


@attr.s(auto_attribs=True)
class Style:
    font: Font = Font()
    border: Border = Border()
    fill: Fill = Fill()
    alignment: Alignment = Alignment()

    @classmethod
    def from_json(cls, style_json: Dict, default_style: 'Style' = None) -> 'Style':
        default_style = default_style or Style()

        font_data = style_json.get("font", {})
        font_data = {**vars(default_style.font), **font_data}
        font = Font(**font_data)

        sides_data = style_json.get("border", {})
        border_data = {side: Side(**side_data) for side, side_data in sides_data.items()}
        border_data = {**vars(default_style.border), **border_data}
        border = Border(**border_data)

        fill_data = style_json.get("fill", {})
        fill_data = {**vars(default_style.fill), **fill_data}
        fill = PatternFill(**fill_data)

        alignment_data = style_json.get("alignment", {})
        alignment_data = {**vars(default_style.alignment), **alignment_data}
        alignment = Alignment(**alignment_data)

        return cls(font, border, fill, alignment)


@attr.s(auto_attribs=True)
class CellWithSize:
    cell: Cell
    width: int
    height: int
